import sys
sys.path.append('../../')
from OrthNet.orthnet.utils import Poly, Poly1d
import tensorflow as tf
import torch
from torch.autograd import Variable
import numpy as np

tf_initial = [lambda x: tf.ones_like(x), lambda x: x]
tf_recurrence = lambda p1, p2, n, x: ((2*n+1)*tf.multiply(x, p1)-n*p2)/(n+1)
tf_x = tf.placeholder(dtype = tf.float32)
tf_legendre = Poly1d('tensorflow', 3, tf_x, tf_initial, tf_recurrence)
tf_x_data = np.linspace(-1, 1, 5).reshape((-1, 1))
with tf.Session() as sess:
	y1 = sess.run([tf_legendre.tensor], feed_dict = {tf_x: tf_x_data})[0]
print(y1)
print(tf_legendre.degree)
tf_legendre.update(4)
with tf.Session() as sess:
	y = sess.run([tf_legendre.tensor], feed_dict = {tf_x: tf_x_data})[0]
print(y)
print(tf_legendre.degree)

# x1_data = np.linspace(-1, 1, 5).reshape((-1, 1))
# x1 = Variable(torch.Tensor(x1_data))
# initial = [lambda x: torch.ones_like(x), lambda x: x]
# recurrence = lambda p1, p2, n, x: ((2*n+1)*x*p1-n*p2)/(n+1)
# y1 = Poly1d('pytorch', 3, x1, initial, recurrence)
# print(y1.tensor)
# y1.update(4)
# print(y1.tensor)


import tensorflow as tf
import torch
import numpy as np
from ..utils import Poly

class Laguerre(Poly):
	"""
	Laguerre Polynomials
	"""
	def __init__(self, module, degree, x, dtype = 'float32', loglevel = 0):
		"""
		input:
			module: 'tensorflow', 'pytorch' or 'numpy'
			degree: highest degree of polynomial
			x: a tensor of shape [Nsample*Nparameter], each row is a sample point, each column represents a parameter
			dtype: 'float32' or 'float64'
			loglevel: 1 to print time info and 0 to mute
		"""
		if module == 'tensorflow':
			initial = [lambda x: tf.ones_like(x), lambda x: 1-x]
			recurrence = lambda p1, p2, n, x: (tf.multiply(2*n+1-x, p1)-n*p2)/(n+1)
		elif module == 'pytorch':
			initial = [lambda x: torch.ones_like(x), lambda x: 1-x]
			recurrence = lambda p1, p2, n, x: ((2*n+1-x)* p1-n*p2)/(n+1)
		elif module == 'numpy':
			initial = [lambda x: np.ones_like(x), lambda x: 1-x]
			recurrence = lambda p1, p2, n, x: ((2*n+1-x)* p1-n*p2)/(n+1)
		Poly.__init__(self, module, degree, x, initial, recurrence, dtype, loglevel)


import tensorflow as tf
import torch
from multi_dim import enumerate_dim as enumerate_dim
from _enum_dim import enum_dim as enum_dim_cpp


class Poly1d:
	"""
	Base class, 1-dimensional orthogonal polynomials by three-term recursion.
	"""
	def __init__(self, module, degree, x, initial, recurrence):
		"""
		- input:
			- module: 'tensorflow' or 'pytorch' (case insensitive)
			- degree: degree of polynomial
			- x: tensor for evaluating function values (should have dimension >= 2)
			- initial: initial value, a list of two functions
			- recurrence: the function of recurrence, with three variables:
				P_{n+1} = f(P_{n}, P_{n-1}, n, x) 
			(initial and recurrence are functions on Tensors)
		"""
		self.module = module.lower()
		assert self.module in ['tensorflow', 'pytorch'], "Module should be either 'tensorflow' or 'pytorch'."
		assert n >= 0 and isinstance(n, int), "Degree should be a non-negative integer."
		assert len(initial) == 2, "Need two initial functions."
		if self.module == 'tensorflow':
			assert isinstance(x, tf.Variable) or isinstance(x, tf.Tensor), "x should be an isinstance of tensorflow.Variable or tensorflow.Tensor."
		else:
			assert isinstance(x, torch.autograd.Variable) or isinstance(x, torch.Tensor), "x should be an isinstance of torch.autograd.Variable or torch.Tensor."
		self.degree = degree
		self.x = x
		self.initial = initial
		self.recurrence = recurrence
		self.poly_list = [self.initial[0](self.x), self.initial[1](self.x)]

	def _compute(self, start, end):
		if end == 0:
			return [self.initial[0](self.x)]
		elif end == 1:
			return [self.initial[0](self.x), self.initial[1](self.x)]
		else:
			polys = []
			for i in range(start, end+1):

	def update(self, new_degree):
		self.
				

	@property
	def list(self):
		"""
		generate a list of function values in lexicographical order.

		output:
			[[p0(x1), p0(x2), .., p0(xm)], [p1(x1), p1(x2), .., p1(xm)], .., [pn(x1), pn(x2), .., pn(xm)]]
		"""
		if self.poly_list:
			return self.poly_list
		if self.order == 0:
			self.poly_list = [initial[0](self.x)]
		elif self.order == 1:
			self.poly_list = [initial[0](x), initial[1](x)]
		else: 
			self.poly_list = [initial[0](x), initial[1](x)]
			for i in range(1, n):
				self.poly_list.append(self.recurrence(self.poly_list[-1], self.poly_list[-2], i, self.x))
			return self.poly_list

	@property
	def tensor(self):
		"""
		generate a tensor of function values in lexicographical order.

		output:
			Tensor([[p0(x)], [p1(x)], .., [pn(x)]])
		"""
		if self.module == 'tensorflow':
			return tf.concat(self.polys_list, axis = 1)
		else:
			return torch.cat(self.polys_list, dim = 1)



class Poly:
	"""
	Base class, multi-dimensional orthogonal polynomials by three-term recursion and tensor product.
	"""
	def __init__(self, module, n, x, initial, recurrence):
		"""
		input:
			- module: 'tensorflow' or 'pytorch'
			- n: order of target polynomial
			- x: a tensor, each row is a sample point, and each column is a feature (or variable).
			- initial: initial value, a list of two functions
			- recurrence: the function of recurrence, with three variables:
				P_{n+1} = f(P_{n}, P_{n-1}, n, x) 
			(initial and recurrence are functions on Tensors)

		output:
			- y: a list of function values
		"""
		self.module = module.lower()
		assert self.module in ['tensorflow', 'pytorch'], "Module should be either 'tensorflow' or 'pytorch'."
		assert n >= 0 and isinstance(n, int), "Order should be a non-negative integer."
		assert len(initial) == 2, "Need two initial functions."
		self.n = n
		self.x = x
		self.initial = initial
		self.recurrence = recurrence
		self.dims = []
		if self.module == 'tensorflow':
			self.n_features

	@property
	def list(self):
		"""
		generate a list of half tensor product of function values in lexicographical order.

		example:
			>>> x = Poly(_, 2, [x, y], _)
			>>> x.polys_list
			>>> [[p0(x)p0(y)], [p1(x)p0(y)], [p0(x)p1(y)], [p2(x)p0(y)], [p1(x)p1(y)], [p0(x)p2(y)]]
		"""
		one_dim_polys, polys = [], []
		if 
		# for index in range()
		# 	tmp = Poly1d(self.module, self.n, var, self.initial, self.recurrence)
		# 	one_dim_polys.append(tmp.list)
		if not self.dims:
			for i in range(n+1):
				self.dims.append(enumerate_dim_cpp())
			



# class Poly(Poly1d):
# 	"""
# 	Base class, multi-dimensional orthogonal polynomials by three-term recursion and tensor product.
# 	"""
# 	def __init__(self, module, n, x, initial, recurrence):
# 		"""
# 		input:
# 			- module: 'tensorflow' or 'pytorch'
# 			- n: order of target polynomial
# 			- x: a list of tensors (as variables)
# 			- initial: initial value, a list of two functions
# 			- recurrence: the function of recurrence, with three variables:
# 				P_{n+1} = f(P_{n}, P_{n-1}, n, x) 
# 			(initial and recurrence are functions on Tensors)

# 		output:
# 			- y: a list of function values
# 		"""
# 		assert isinstance(x, list) or isinstance(x, tuple), "x should be a list or a tuple of tensors."
# 		super().__init__(self, module, n, x[0], initial, recurrence)
# 		self.x = x # Override super.x

# 	@property
# 	def list(self):
# 		"""
# 		generate a list of half tensor product of function values in lexicographical order.

# 		example:
# 			>>> x = Poly(_, 2, [x, y], _)
# 			>>> x.polys_list
# 			>>> [[p0(x)p0(y)], [p1(x)p0(y)], [p0(x)p1(y)], [p2(x)p0(y)], [p1(x)p1(y)], [p0(x)p2(y)]]
# 		"""
# 		one_dim_polys, polys = [], []
# 		for var in self.x:
# 			tmp = Poly1d(self.module, n, x, initial, recurrence)
# 			one_dim_polys.append()


import tensorflow as tf
import torch
from ..utils import Poly1d, Poly
import numpy as np

class Legendre(Poly):
	"""
	Legendre Polynomials
	"""
	def __init__(self, module, degree, x, dtype = 'float32', loglevel = 0):
		"""
		input:
			module: 'tensorflow', 'pytorch', or 'numpy'
			degree: highest degree of polynomial
			x: a tensor of shape [Nsample*Nparameter], each row is a sample point, each column represents a parameter
			dtype: 'float32' or 'float64'
			loglevel: 1 to print time info and 0 to mute
		"""

		if module == 'tensorflow':
			initial = [lambda x: tf.ones_like(x), lambda x: x]
			recurrence = lambda p1, p2, n, x: ((2*n+1)*tf.multiply(x, p1)-n*p2)/(n+1)
		elif module == 'pytorch':
			initial = [lambda x: torch.ones_like(x), lambda x: x]
			recurrence = lambda p1, p2, n, x: ((2*n+1)*x*p1-n*p2)/(n+1)
		elif module == 'numpy':
			initial = [lambda x: np.ones_like(x), lambda x: x]
			recurrence = lambda p1, p2, n, x: ((2*n+1)*x*p1-n*p2)/(n+1)
		Poly.__init__(self, module, degree, x, initial, recurrence, dtype, loglevel)


class Legendre_Normalized(Poly):
	"""
	Normalized Legendre Polynomials with integral be 1 if n = m.
	"""
	def __init__(self, module, degree, x, dtype = 'float32', loglevel = 0):
		"""
		input:
			module: 'tensorflow' or 'pytorch'
			degree: highest degree of polynomial
			x: a tensor of shape [Nsample*Nparameter], each row is a sample point, each column represents a parameter
			dtype: 'float32' or 'float64'
			loglevel: 1 to print time info and 0 to mute
		"""

		if module == 'tensorflow':
			initial = [lambda x: tf.ones_like(x)*np.sqrt(1/2), lambda x: x*np.sqrt(3/2)]
			recurrence = lambda p1, p2, n, x: (np.sqrt((2*n+1)*(2*n+3))*tf.multiply(x, p1)-n*np.sqrt((2*n+3)/(2*n-1))*p2)/(n+1)
		elif module == 'pytorch':
			initial = [lambda x: torch.ones_like(x)*np.sqrt(1/2), lambda x: x*np.sqrt(3/2)]
			recurrence = lambda p1, p2, n, x: (np.sqrt((2*n+1)*(2*n+3))*x*p1-n*np.sqrt((2*n+3)/(2*n-1))*p2)/(n+1)
		elif module == 'numpy':
			initial = [lambda x: np.ones_like(x)*np.sqrt(1/2), lambda x: x*np.sqrt(3/2)]
			recurrence = lambda p1, p2, n, x: (np.sqrt((2*n+1)*(2*n+3))*x*p1-n*np.sqrt((2*n+3)/(2*n-1))*p2)/(n+1)
		Poly.__init__(self, module, degree, x, initial, recurrence, dtype, loglevel)


import sys
sys.path.append('../../')
from OrthNet.orthnet.poly.legendre import Legendre
import tensorflow as tf
import torch
from torch.autograd import Variable
import numpy as np

N = 3
x1 = tf.placeholder(dtype = tf.float32, shape = [None, N])
l1 = Legendre('tensorflow', 2, x1)
x2 = tf.placeholder(dtype = tf.float32, shape = [None, 1])
l2 = Legendre('tensorflow', 2, x2)
with tf.Session() as sess:
	x1_data = np.linspace(-1, 1, N)
	x1_data = np.asarray([x1_data, x1_data])
	y1 = sess.run(l1.tensor, feed_dict = {x1:x1_data})
	print(sess.run(l1.get_poly(2, 2), feed_dict = {x1:x1_data}))


print(y1)
print(l1.combination)
print(l1.length)


import tensorflow as tf
import torch
import numpy as np
from ._enum_dim import enum_dim as enum_dim
from .timeit import timeit
import time
import pickle


class Poly1d:
	"""
	1-dimensional orthogonal polynomials by three-term recursion.
	"""
	def __init__(self, module, degree, x, initial, recurrence, dtype = 'float32'):
		"""
		- input:
			- module: 'tensorflow', 'pytorch' or 'numpy' (case insensitive)
			- degree: degree of polynomial
			- x: a tensor of shape (n*1), where `n` is the number of sample points
			- initial: initial value, a list of two functions
			- recurrence: the function of recurrence, with three variables:
				P_{n+1} = f(P_{n}, P_{n-1}, n, x) 
			(initial and recurrence are functions on Tensors)
			- dtype: 'float32' or 'float64'
		"""
		self.module = module.lower()
		assert self.module in ['tensorflow', 'pytorch', 'numpy'], "Module should be one of ['tensorflow', 'pytorch', 'numpy']."
		assert degree >= 0 and isinstance(degree, int), "Degree should be a non-negative integer."
		assert len(initial) == 2, "Need two initial functions."
		if self.module == 'tensorflow':
			assert isinstance(x, tf.Variable) or isinstance(x, tf.Tensor), "x should be an isinstance of tensorflow.Variable or tensorflow.Tensor."
		elif self.module == 'pytorch':
			assert isinstance(x, torch.autograd.Variable) or isinstance(x, torch.Tensor) or isinstance(x, torch.DoubleTensor), "x should be an isinstance of torch.autograd.Variable or torch.Tensor."
		else:
			assert isinstance(x, np.ndarray) or isinstance(x, np.matrix), "x should be an isinstance of numpy.array or numpy.matrix"
		assert dtype in ['float32', 'float64'], "dtype should be either 'float32' or 'float64'"
		self.degree = degree
		self.x = x
		self.recurrence = recurrence
		self.dtype = dtype
		self._list = [initial[0](self.x), initial[1](self.x)]

	def _compute(self, start, end):
		"""
		Compute polynomials from degree `start`(included) to `end`(included)
		"""
		for i in range(start-1, end):
			self._list.append(self.recurrence(self._list[-1], self._list[-2], i, self.x))

	@property
	def list(self):
		"""
		return a list of polynomials
		"""
		self._compute(len(self._list), self.degree)
		return [self._list[0]] if self.degree == 0 else self._list

	@property
	def tensor(self):
		"""
		return a tensor of polynomials
		"""
		if self.module == 'tensorflow':
			return tf.transpose(tf.concat(self.list, axis = 1))
		elif self.module == 'pytorch':
			return torch.transpose(torch.cat(self.list, dim = 1))
		else:
			return np.transpose(np.concatenate(self.list, axis = 1))

	def update(self, newdegree):
		"""
		update polynomials to a higher degree
		"""
		self._compute(self.degree+1, newdegree)
		self.degree = newdegree

	def quadrature(self, func, weight):
		if self.module == 'tensorflow':
			weight = tf.reshape(tf.constant(weight), shape = [1, -1])
			return tf.matmul(weight, tf.multiply(func(self.x), self.tensor))
		elif self.module == 'pytorch':
			weight = torch.tensor(weight, shape = [1, -1])
			return torch.matmul(weight, func(self.x)*self.tensor)
		else:
			weight = weight.reshape((1, -1))
			return np.dot(weight, func(self.x)*self.tensor)



class Poly:
	"""
	n-dimensional orthogonal polynomials by three-term recursion and tensor product
	"""
	def __init__(self, module, degree, x, initial, recurrence, dtype = 'float32', loglevel = 0):
		"""
		- input:
			- module: 'tensorflow', 'pytorch' or 'numpy' (case insensitive)
			- degree: degree of polynomial
			- x: a tensor of shape (num_sample*num_parameter)
			- initial: initial value, a list of two functions
			- recurrence: the function of recurrence, with three variables:
				P_{n+1} = f(P_{n}, P_{n-1}, n, x) 
			(initial and recurrence are functions on Tensors)
			- dtype: 'float32' or 'float64'
		"""
		self.module = module.lower()
		assert self.module in ['tensorflow', 'pytorch', 'numpy'], "Module should be one of ['tensorflow', 'pytorch', 'numpy']."
		assert degree >= 0 and isinstance(degree, int), "Degree should be a non-negative integer."
		assert len(initial) == 2, "Need two initial functions."
		if self.module == 'tensorflow':
			assert isinstance(x, tf.Variable) or isinstance(x, tf.Tensor), "x should be an isinstance of tensorflow.Variable or tensorflow.Tensor."
		elif self.module == 'pytorch':
			assert isinstance(x, torch.autograd.Variable) or isinstance(x, torch.Tensor) or isinstance(x, torch.DoubleTensor), "x should be an isinstance of torch.autograd.Variable or torch.Tensor."
		else:
			assert isinstance(x, np.ndarray) or isinstance(x, np.matrix), "x should be an isinstance of numpy.array or numpy.matrix"
		assert dtype in ['float32', 'float64'], "dtype should be either 'float32' or 'float64'"
		self.degree = degree
		self.x = x
		self.initial = initial
		self.recurrence = recurrence
		self.dtype = dtype
		self.loglevel = loglevel
		if self.module == 'tensorflow':
			self.dim = self.x.get_shape()[1].value
		elif self.module == 'pytorch':
			self.dim = x.size()[1]
		else:
			self.dim = x.shape[1]
		self._init()
		self._list = []	

	def _init(self):
		self._comb()
		self._poly1d = [Poly1d(self.module, self.degree, self.x[:, i], self.initial, self.recurrence, self.dtype) for i in range(self.dim)]

	def _comb(self):
		@timeit(self.loglevel)
		def __comb(self):
			comb_file_name = "comb_"+str(self.degree)+"_"+str(self.dim)
			try:
				comb = pickle.load(open(comb_file_name, "rb"))
			except FileNotFoundError:
				comb = enum_dim(self.degree, self.dim)
				pickle.dump(comb, open(comb_file_name, "wb"))
			except EOFError:
				comb = enum_dim(self.degree, self.dim)
				pickle.dump(comb, open(comb_file_name, "wb"))
			self._index = comb[0]
			self._combination = comb[1:]
		__comb(self)

	def _compute(self, start, end):
		"""
		compute polynomials from degree `start`(included) and `end`(included).
		"""
		@timeit(self.loglevel)
		def __compute(self, start, end):
			if end == self.degree:
				comb = self._combination[self._index[start]:]
			else:
				comb = self._combination[self._index[start]:self._index[end]+1]
			res = []
			for c in comb:
				if self.module == 'tensorflow':
					if self.dtype == 'float32':
						poly = tf.constant(1, dtype = tf.float32)
					else:
						poly = tf.constant(1, dtype = tf.float64)
					for i in range(len(c)):	
						poly = tf.multiply(poly, self._poly1d[i].list[c[i]])
					poly = tf.reshape(poly, [-1, 1])
				elif self.module == 'pytorch':
					if self.dtype == 'float32':
						poly = torch.ones([self.x.size()[0]])
					else:
						poly = torch.ones([self.x.size()[0]]).double()
					for i in range(len(c)):
						poly = poly*self._poly1d[i].list[c[i]]
					poly.unsqueeze_(1)
				else:
					if self.dtype == 'float32':
						poly = np.ones([self.x.shape[0]], dtype = np.float32)
					else:
						poly = np.ones([self.x.shape[0]], dtype = np.float64)
					for i in range(len(c)):
						poly = poly*self._poly1d[i].list[c[i]]
					poly = np.expand_dims(poly, axis = 1)
				res.append(poly)
			return res
		return __compute(self, start, end)

	@property
	def length(self):
		"""
		return the number of polynomials
		"""
		return len(self.combination)

	@property
	def list(self):
		"""
		return a list of polynomials
		"""
		if not self._list:
			self._list.extend(self._compute(0, self.degree))
		return self._list

	@property
	def tensor(self):
		"""
		return a tensor of polynomials
		"""
		if self.module == 'tensorflow':
			return tf.concat(self.list, axis = 1)
		elif self.module == 'pytorch':
			return torch.cat(self.list, dim = 1)
		else:
			return np.concatenate(self.list, axis = 1)

	def update(self, newdegree):
		"""
		update to a higher degree
		"""
		if newdegree > self.degree:
			original_degree = self.degree
			self.degree = newdegree
			self._comb()
			for i in range(self.dim):
				self._poly1d[i].update(newdegree)
			self._list.extend(self._compute(original_degree+1, newdegree))

	@property
	def index(self):
		"""
		return the index of the first combination of each degree
		"""
		return self._index

	@property
	def combination(self):
		"""
		return all combinations of all degrees
		"""
		return self._combination

	def get_one_degree_combination(self, degree):
		"""
		return all combination of a given degree
		"""
		assert degree <= self.degree, "Degree should be less or equal than the highest degree."
		if degree == self.degree:
			return self._combination[self._index[degree]:]
		else:
			return self._combination[self._index[degree]:self._index[degree+1]]

	def get_combination(self, start, end):
		"""
		return all combination of degrees from `start`(included)  to `end`(included)
		"""
		res = []
		for degree in range(start, end+1):
			res.extend(self.get_one_degree_combination(degree))
		return tuple(res)

	def get_poly(self, start, end):
		"""
		get polynomials from degree `start`(included) to `end`(included)
		"""
		assert start >= 0 and end <= self.degree, "Degree should be less or equal than highest degree"
		if end == self.degree:
			return self.tensor[:, self._index[start]:]
		else:
			return self.tensor[:, self._index[start]:self._index[end+1]]

	def eval(self, coeff):
		"""
		return the values with coefficients given
		"""
		assert len(coeff) == self.length, "Coefficients should have a length equal to number of polynomials"
		if self.module == 'tensorflow':
			coeff = tf.reshape(tf.constant(coeff), shape = [-1, 1])
			return tf.matmul(self.tensor, coeff)
		elif self.module == 'pytorch':
			coeff = torch.Tensor(coeff, shape = [-1, 1])
			return torch.matmul(self.tensor, coeff)
		else:
			coeff = coeff.reshape((-1, 1))
			return np.dot(self.tensor, coeff)

	def quadrature(self, func, weight):
		if self.module == 'tensorflow':
			weight = tf.reshape(tf.constant(weight), shape = [1, -1])
			return tf.matmul(weight, tf.multiply(func(self.x), self.tensor))
		elif self.module == 'pytorch':
			weight = torch.Tensor(weight).view(1, -1)
			return torch.matmul(weight, func(self.x)*self.tensor)
		else:
			weight = weight.reshape((1, -1))
			return np.dot(weight, func(self.x)*self.tensor)


import time
from functools import wraps

def timeit(loglevel):
	def _timeit(func):
		@wraps(func)
		def timed(*args, **kw):
			t1 = time.time()
			res = func(*args, **kw)
			print(func.__name__, (time.time() - t1))
			return res
		def untimed(*args, **kw):
			return func(*args, **kw)
		if loglevel == 0:
			return untimed
		elif loglevel == 1:
			return timed
	return _timeit


import sys
sys.path.append('..')
from orthnet import Laguerre, Legendre, Legendre_Normalized, Chebyshev, Chebyshev2, Hermite, Hermite2, Jacobi
import numpy as np
import tensorflow as tf
import torch


def integral(f1, f2, x, weight):
	"""
	compute numerical integration with f1, f2, weight, and x

	input:
		f1, f2: vectors of function values
		x: points for integration 
		weight: the weight function
	"""
	return np.trapz(f1*f2*weight(x), x = x)

def test(degree, x, y, weight):
	res = np.zeros((degree+1, degree+1))
	for i in range(degree+1):
		for j in range(degree+1):
			res[i, j] = integral(y[:, i], y[:, j], x, weight)
	return res

def test_legendre(degree):
	x = np.linspace(-1, 1, 2, dtype = np.float32)
	x_tf = tf.reshape(tf.constant(x), [-1, 1])
	l = Legendre_Normalized('tensorflow', degree, x_tf)
	weight = lambda x: 1
	with tf.Session() as sess:
		y = sess.run(l.tensor)
	return test(degree, x, y, weight)

def test_laguerre(degree):
	x = np.linspace(0, 1e3, 10000, dtype = np.float32)
	x_tf = tf.reshape(tf.constant(x), [-1, 1])
	l = Laguerre('tensorflow', degree, x_tf)
	weight = lambda x: np.exp(-x)
	with tf.Session() as sess:
		y = sess.run(l.tensor)
	return test(degree, x, y, weight)

def test_hermite(degree):
	x = np.linspace(-1e4, 1e4, 100000, dtype = np.float32)
	x_tf = tf.reshape(tf.constant(x), [-1, 1])
	l = Hermite('tensorflow', degree, x_tf)
	weight = lambda x: np.exp(-x**2/2)
	with tf.Session() as sess:
		y = sess.run(l.tensor)
	return test(degree, x, y, weight)

def test_hermite2(degree):
	x = np.linspace(-1e4, 1e4, 100000, dtype = np.float32)
	x_tf = tf.reshape(tf.constant(x), [-1, 1])
	l = Hermite2('tensorflow', degree, x_tf)
	weight = lambda x: np.exp(-x**2)
	with tf.Session() as sess:
		y = sess.run(l.tensor)
	return test(degree, x, y, weight)

def test_chebyshev(degree):
	x = np.linspace(-1+1e-6, 1-1e-6, 100000, dtype = np.float32)
	x_tf = tf.reshape(tf.constant(x), [-1, 1])
	l = Chebyshev('tensorflow', degree, x_tf)
	weight = lambda x: 1./np.sqrt(1.-x**2)
	with tf.Session() as sess:
		y = sess.run(l.tensor)
	return test(degree, x, y, weight)

def test_chebyshev2(degree):
	x = np.linspace(-1+1e-6, 1-1e-6, 100000, dtype = np.float32)
	x_tf = tf.reshape(tf.constant(x), [-1, 1])
	l = Chebyshev2('tensorflow', degree, x_tf)
	weight = lambda x: np.sqrt(1.-x**2)
	with tf.Session() as sess:
		y = sess.run(l.tensor)
	return test(degree, x, y, weight)

def test_jacobi(degree, alpha, beta):
	x = np.linspace(-1+1e-6, 1-1e-6, 100000, dtype = np.float32)
	x_tf = tf.reshape(tf.constant(x), [-1, 1])
	l = Jacobi('tensorflow', degree, x_tf, alpha, beta)
	weight = lambda x: ((1-x)**alpha)*((1+x)**beta)
	with tf.Session() as sess:
		y = sess.run(l.tensor)
	return test(degree, x, y, weight)



if __name__ == '__main__':
	print(test_legendre(2))
	print(test_laguerre(2))
	print(test_hermite(2))
	print(test_hermite2(2))
	print(test_chebyshev(2))
	print(test_chebyshev2(2))
	print(test_jacobi(2, 1, 3))


import tensorflow as tf
import torch
import numpy as np
from ..utils import Poly1d, Poly

class Hermite(Poly):
	"""
	Hermite Polynomials of the first kind (in probability theory)
	"""
	def __init__(self, module, degree, x, dtype = 'float32', loglevel = 0):
		"""
		input:
			module: 'tensorflow', 'pytorch' or 'numpy'
			degree: highest degree of polynomial
			x: a tensor of shape [Nsample*Nparameter], each row is a sample point, each column represents a parameter
			dtype: 'float32' or 'float64'
			loglevel: 1 to print time info and 0 to mute
		"""

		if module == 'tensorflow':
			initial = [lambda x: tf.ones_like(x), lambda x: x]
			recurrence = lambda p1, p2, n, x: tf.multiply(x, p1) - n*p2
		elif module == 'pytorch':
			initial = [lambda x: torch.ones_like(x), lambda x: x]
			recurrence = lambda p1, p2, n, x: x*p1 - n*p2
		elif module == 'numpy':
			initial = [lambda x: np.ones_like(x), lambda x: x]
			recurrence = lambda p1, p2, n, x: x*p1 - n*p2			
		Poly.__init__(self, module, degree, x, initial, recurrence, dtype, loglevel)
		

class Hermite2(Poly):
	"""
	Hermite Polynomials of the second kind (in Physics)
	"""
	def __init__(self, module, degree, x, dtype = 'float32', loglevel = 0):
		"""
		input:
			module: 'tensorflow', 'pytorch' or 'numpy'
			degree: highest degree of polynomial
			x: a tensor of shape [Nsample*Nparameter], each row is a sample point, each column represents a parameter
			dtype: 'float32' or 'float64'
			loglevel: 1 to print time info and 0 to mute
		"""

		if module == 'tensorflow':
			initial = [lambda x: tf.ones_like(x), lambda x: 2*x]
			recurrence = lambda p1, p2, n, x: 2*(tf.multiply(x, p1) - n*p2)
		elif module == 'pytorch':
			initial = [lambda x: torch.ones_like(x), lambda x: 2*x]
			recurrence = lambda p1, p2, n, x: 2*(x*p1 - n*p2)
		elif module == 'numpy':
			initial = [lambda x: np.ones_like(x), lambda x: 2*x]
			recurrence = lambda p1, p2, n, x: 2*(x*p1 - n*p2)
		Poly.__init__(self, module, degree, x, initial, recurrence, dtype, loglevel)


def enumerate_dim(n, dim):
	"""
	enumerate dims:
		find DIM nonnegative numbers, the sum of which is n, return result in lexicographical order.

	input:
		n: total order
		dim: dimension (number of variables)

	output:
		a 2D-list, each element a combination (a list)

	>>> enumerate_dim(3, 2)
	>>> [[3, 0], [2, 1], [1, 2], [0, 3]]
	"""
	def dfs(res, cur, n, dim):
		if dim == 1:
			res.append(cur+[n])
			return
		for i in reversed(range(n+1)):
			dfs(res, cur+[i], n-i, dim-1)
	res = []
	dfs(res, [], n, dim)
	# print('Order:', res)
	return res

def enum_dim(n, dim):
	"""
	enumerate dims:
		find DIM nonnegative numbers, the sum of which <= n, return result in lexicographical order.

	input:
		n: total order
		dim: dimension (number of variables)

	output:
		a 2D-list, each element a combination (a list)

	>>> enum_dim(3, 2)
	>>> [[0, 0], [1, 0], [0, 1], [2, 0], [1, 1], [0, 2], [3, 0], [2, 1], [1, 2], [0, 3]]
	"""
	res = [[[0 for i in range(dim)]]]
	for i in range(n):
		cur = []
		for comb in res[-1]:
			for j in range(len(comb)):
				tmp = comb[:]
				tmp[j] += 1
				flag = 1
				for k in cur:
					if tmp == k:
						flag = 0
						break
				if flag:
					cur.append(tmp)
		res.append(cur)
	return res


def dim(n, d):
	res = []
	for i in range(n+1):
		res.append(enum_dim(i, d))
	return res


from _enum_dim import enum_dim as enum_dim_cpp
from multi_dim import enum_dim as enum_dim_recursive
from multi_dim import dim as enum_dim_py

for i in [3, 5, 7, 9]:
	for j in [10, 20, 30, 40, 50]:
		print(i, j)
		%timeit enum_dim_cpp(i, j)
		%timeit enum_dim_py(i, j)       
		%timeit enum_dim_recursive(i, j)



import sys
sys.path.append('../')
from orthnet import Laguerre
import tensorflow as tf
import torch
from torch.autograd import Variable
import numpy as np


x_data = np.random.random((3, 2))
x_tf = tf.placeholder(dtype = tf.float32, shape = (None, 2))
x_torch = Variable(torch.Tensor(x_data))

y_tf = Laguerre('tensorflow', 2, x_tf)
y_torch = Laguerre('pytorch', 2, x_torch)

with tf.Session() as sess:
	z_tf = sess.run(y_tf.tensor, feed_dict = {x_tf:x_data})
z_torch = y_torch.tensor.data.numpy()

print('length:', y_tf.length)
print('combination:', y_tf.combination)
print('index:', y_tf.index)
print('tf values:')
print(z_tf)
print('torch values:')
print(z_torch)
print('equal?')
print(np.allclose(z_tf, z_torch))
print('\n')

y_tf.update(4)
y_torch.update(4)

with tf.Session() as sess:
	z_tf = sess.run(y_tf.tensor, feed_dict = {x_tf:x_data})
z_torch = y_torch.tensor.data.numpy()

print('updated:')
print('length:', y_tf.length)
print('combination:', y_tf.combination)
print('index:', y_tf.index)
print('tf values:')
print(z_tf)
print('torch values:')
print(z_torch)
print('equal?')
print(np.allclose(z_tf, z_torch))



from distutils.core import setup, Extension

ext = Extension('orthnet.utils._enum_dim',
	sources = ['orthnet/utils/enum_dim/enum_dim.i'],
	language = 'c++',
	swig_opts = ['-c++'],
	extra_compile_args = [
		'-std=c++11',
		'-Wall',
		'-Ofast',
	]
)

setup(
	name = 'orthnet',
	version = '0.3.0',
	keywords = ['orthogonal polynomial', 'tensorflow', 'pytorch', 'numpy'],
	description = 'TensorFlow, PyTorch and Numpy layers for generating orthogonal polynomials',
	license = 'MIT',
	author = 'Chuan Lu',
	author_email = 'chuan-lu@uiowa.edu',
	ext_modules = [ext],
	packages = ['orthnet', 'orthnet.poly', 'orthnet.utils'],
	)


from .legendre import Legendre, Legendre_Normalized
from .laguerre import Laguerre
from .hermite import Hermite, Hermite2
from .chebyshev import Chebyshev, Chebyshev2
from .jacobi import Jacobi

__all__ = ['Legendre', 'Legendre_Normalized', 'Laguerre', 'Hermite', 'Hermite2', 'Chebyshev', 'Chebyshev2', 'Jacobi']


import sys
sys.path.append('../../')
from OrthNet.orthnet.poly.jacobi import Jacobi
from OrthNet.orthnet.poly.legendre import Legendre
import tensorflow as tf
import torch
from torch.autograd import Variable
import numpy as np

N = 3
x1 = tf.placeholder(dtype = tf.float32, shape = [None, N])
l1 = Jacobi('tensorflow', 2, x1, 0, 0)
l2 = Legendre('tensorflow', 2, x1)

with tf.Session() as sess:
	x1_data = np.linspace(-1, 1, N)
	x1_data = np.asarray([x1_data, x1_data])
	y1 = sess.run(l1.tensor, feed_dict = {x1:x1_data})
	y2 = sess.run(l2.tensor, feed_dict = {x1:x1_data})


print(y1)
print(y2)


import sys
sys.path.append('../../')
from OrthNet.orthnet.utils.poly import Poly
import tensorflow as tf
import torch
from torch.autograd import Variable
import numpy as np

tf_initial = [lambda x: tf.ones_like(x), lambda x: x]
tf_recurrence = lambda p1, p2, n, x: ((2*n+1)*tf.multiply(x, p1)-n*p2)/(n+1)
tf_x = tf.placeholder(dtype = tf.float32, shape=[None, 3])
tf_legendre = Poly('tensorflow', 1, tf_x, tf_initial, tf_recurrence)
tf_x_data = np.linspace(-1, 1, 6).reshape((-1, 3))

with tf.Session() as sess:
	y = sess.run(tf_legendre.tensor, feed_dict = {tf_x: tf_x_data})
print(tf_legendre.length)

tf_legendre.update(4)
with tf.Session() as sess:
	print(sess.run(tf_legendre.tensor, feed_dict = {tf_x: tf_x_data}))
print(tf_legendre.index)
print(tf_legendre.combination)

# import time
# t1 = time.time()
# x = tf_legendre.combination
# t2 = time.time()
# x = tf_legendre.combination
# t3 = time.time()
# print(x)

# print('t1:', t2-t1)
# print('t2:', t3-t2)